import numpy as np
from mattress import initialize

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_dense():
    for dt in [np.float64, np.float32, np.int64, np.int32, np.int16, np.int8, np.uint64, np.uint32, np.uint16, np.uint8]:
        y = (np.random.rand(1000, 100) * 100).astype(dt, order="C")
        ptr = initialize(y)
        assert all(ptr.row(0) == y[0, :])
        assert all(ptr.column(1) == y[:, 1])
        assert ptr.shape == (1000, 100)
        assert ptr.dtype == np.float64


def test_dense_column_major():
    y = np.ndarray((1000, 100), order="F")
    y[:, :] = np.random.rand(1000, 100)
    assert y.flags["F_CONTIGUOUS"]
    ptr = initialize(y)
    assert all(ptr.row(0) == y[0, :])
    assert all(ptr.column(1) == y[:, 1])



import numpy as np
import delayedarray as da
from mattress import initialize

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_delayed_unary_isometric_arith_scalar():
    y = np.random.rand(1000, 100)

    # Addition.
    x = da.DelayedArray(y) + 1
    assert isinstance(x.seed, da.UnaryIsometricOpWithArgs)
    ptr = initialize(x)
    ref = y + 1
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = 2 + da.DelayedArray(y)
    ptr = initialize(x)
    ref = y + 2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Subtraction.
    x = da.DelayedArray(y) - 5.5
    ptr = initialize(x)
    ref = y - 5.5
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = 2.3 - da.DelayedArray(y)
    ptr = initialize(x)
    ref = 2.3 - y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Multiplication.
    x = da.DelayedArray(y) * 3.5
    ptr = initialize(x)
    ref = y * 3.5
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = 2.3 * da.DelayedArray(y)
    ptr = initialize(x)
    ref = 2.3 * y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Division.
    x = da.DelayedArray(y) / 2.5
    ptr = initialize(x)
    ref = y / 2.5
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = 2.3 / da.DelayedArray(y)
    ptr = initialize(x)
    ref = 2.3 / y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Modulo.
    x = da.DelayedArray(y) % 0.2
    ptr = initialize(x)
    ref = y % 0.2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = 199 % da.DelayedArray(y)
    ptr = initialize(x)
    ref = 199 % y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Integer division.
    x = da.DelayedArray(y) // 0.1
    ptr = initialize(x)
    ref = y // 0.1
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = 2 // da.DelayedArray(y)
    ptr = initialize(x)
    ref = 2 // y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Power.
    x = da.DelayedArray(y) ** 3.1
    ptr = initialize(x)
    ref = y**3.1
    assert np.allclose(ptr.row(0), ref[0, :])
    assert np.allclose(ptr.column(1), ref[:, 1])

    x = 2 ** da.DelayedArray(y)
    ptr = initialize(x)
    ref = 2**y
    assert np.allclose(ptr.row(0), ref[0, :])
    assert np.allclose(ptr.column(1), ref[:, 1])


def test_delayed_unary_isometric_arith_vector():
    y = np.random.rand(1000, 100)
    v1 = np.random.rand(100)
    v2 = np.random.rand(1000, 1)

    # Addition.
    x = da.DelayedArray(y) + v1
    assert isinstance(x.seed, da.UnaryIsometricOpWithArgs)
    ptr = initialize(x)
    ref = y + v1
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v2 + da.DelayedArray(y)
    ptr = initialize(x)
    ref = y + v2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Subtraction.
    x = da.DelayedArray(y) - v2
    ptr = initialize(x)
    ref = y - v2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v1 - da.DelayedArray(y)
    ptr = initialize(x)
    ref = v1 - y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Multiplication.
    x = da.DelayedArray(y) * v2
    ptr = initialize(x)
    ref = y * v2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v1 * da.DelayedArray(y)
    ptr = initialize(x)
    ref = v1 * y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Division.
    x = da.DelayedArray(y) / v1
    ptr = initialize(x)
    ref = y / v1
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v2 / da.DelayedArray(y)
    ptr = initialize(x)
    ref = v2 / y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Modulo.
    x = da.DelayedArray(y) % v2
    ptr = initialize(x)
    ref = y % v2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v1 % da.DelayedArray(y)
    ptr = initialize(x)
    ref = v1 % y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Integer division.
    x = da.DelayedArray(y) // v1
    ptr = initialize(x)
    ref = y // v1
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v2 // da.DelayedArray(y)
    ptr = initialize(x)
    ref = v2 // y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Power.
    x = da.DelayedArray(y) ** v2
    ptr = initialize(x)
    ref = y**v2
    assert np.allclose(ptr.row(0), ref[0, :])
    assert np.allclose(ptr.column(1), ref[:, 1])

    x = v1 ** da.DelayedArray(y)
    ptr = initialize(x)
    ref = v1**y
    assert np.allclose(ptr.row(0), ref[0, :])
    assert np.allclose(ptr.column(1), ref[:, 1])


def test_delayed_unary_isometric_compare_scalar():
    y = np.random.rand(1000, 100)
    first = float(y[0, 0])

    # Equality.
    x = da.DelayedArray(y) == first
    assert isinstance(x.seed, da.UnaryIsometricOpWithArgs)
    ptr = initialize(x)
    ref = y == first
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = first == da.DelayedArray(y)
    ptr = initialize(x)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Non-equality.
    x = da.DelayedArray(y) != first
    ptr = initialize(x)
    ref = y != first
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = first != da.DelayedArray(y)
    ptr = initialize(x)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Greater than.
    x = da.DelayedArray(y) > first
    ptr = initialize(x)
    ref = y > first
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = first > da.DelayedArray(y)
    ptr = initialize(x)
    ref = first > y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Greater than or equal to.
    x = da.DelayedArray(y) >= first
    ptr = initialize(x)
    ref = y >= first
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = first >= da.DelayedArray(y)
    ptr = initialize(x)
    ref = first >= y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Less than.
    x = da.DelayedArray(y) < first
    ptr = initialize(x)
    ref = y < first
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = first < da.DelayedArray(y)
    ptr = initialize(x)
    ref = first < y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Less than or equal to
    x = da.DelayedArray(y) <= first
    ptr = initialize(x)
    ref = y <= first
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = first <= da.DelayedArray(y)
    ptr = initialize(x)
    ref = first <= y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()


def test_delayed_unary_isometric_compare_vector():
    y = np.random.rand(1000, 100)
    v1 = y[:, [0]]
    v2 = y[0, :]

    # Equality.
    x = da.DelayedArray(y) == v1
    assert isinstance(x.seed, da.UnaryIsometricOpWithArgs)
    ptr = initialize(x)
    ref = y == v1
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v2 == da.DelayedArray(y)
    ptr = initialize(x)
    ref = v2 == y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Non-equality.
    x = da.DelayedArray(y) != v2
    ptr = initialize(x)
    ref = y != v2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v1 != da.DelayedArray(y)
    ref = v1 != y
    ptr = initialize(x)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Greater than.
    x = da.DelayedArray(y) > v1
    ptr = initialize(x)
    ref = y > v1
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v2 > da.DelayedArray(y)
    ptr = initialize(x)
    ref = v2 > y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Greater than or equal to.
    x = da.DelayedArray(y) >= v2
    ptr = initialize(x)
    ref = y >= v2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v1 >= da.DelayedArray(y)
    ptr = initialize(x)
    ref = v1 >= y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Less than.
    x = da.DelayedArray(y) < v2
    ptr = initialize(x)
    ref = y < v2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v1 < da.DelayedArray(y)
    ptr = initialize(x)
    ref = v1 < y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # Less than or equal to
    x = da.DelayedArray(y) <= v2
    ptr = initialize(x)
    ref = y <= v2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = v1 <= da.DelayedArray(y)
    ptr = initialize(x)
    ref = v1 <= y
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()


def test_delayed_unary_isometric_logical_scalar():
    y = np.random.rand(1000, 100)

    # OR.
    x = np.logical_or(da.DelayedArray(y), True)
    assert isinstance(x.seed, da.UnaryIsometricOpWithArgs)
    ptr = initialize(x)
    ref = np.logical_or(y, True)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = np.logical_or(False, da.DelayedArray(y))
    ptr = initialize(x)
    ref = np.logical_or(False, y)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # AND.
    x = np.logical_and(da.DelayedArray(y), False)
    ptr = initialize(x)
    ref = np.logical_and(y, False)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = np.logical_and(True, da.DelayedArray(y))
    ptr = initialize(x)
    ref = np.logical_or(True, y)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # XOR.
    x = np.logical_xor(da.DelayedArray(y), True)
    ptr = initialize(x)
    ref = np.logical_xor(y, True)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = np.logical_xor(False, da.DelayedArray(y))
    ptr = initialize(x)
    ref = np.logical_xor(False, y)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()


def test_delayed_unary_isometric_logical_vector():
    y = np.random.rand(1000, 100)
    v1 = np.random.rand(100) > 0.5
    v2 = np.random.rand(1000, 1) > 0.5

    # OR.
    x = np.logical_or(da.DelayedArray(y), v1)
    assert isinstance(x.seed, da.UnaryIsometricOpWithArgs)
    ptr = initialize(x)
    ref = np.logical_or(y, v1)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = np.logical_or(v2, da.DelayedArray(y))
    ptr = initialize(x)
    ref = np.logical_or(v2, y)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # AND.
    x = np.logical_and(da.DelayedArray(y), v2)
    ptr = initialize(x)
    ref = np.logical_and(y, v2)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = np.logical_and(v1, da.DelayedArray(y))
    ptr = initialize(x)
    ref = np.logical_and(v1, y)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    # XOR.
    x = np.logical_xor(da.DelayedArray(y), v1)
    ptr = initialize(x)
    ref = np.logical_xor(y, v1)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = np.logical_xor(v2, da.DelayedArray(y))
    ptr = initialize(x)
    ref = np.logical_xor(v2, y)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()



from mattress import initialize
import delayedarray
import random
import numpy

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def simulate_SparseNdarray(nr, nc, dtype, itype):
    contents = []
    for i in range(nc):
        if random.random() < 0.5:
            contents.append(None)
            continue
        curdata = []
        curindex = []
        for r in range(nr):
            if random.random() < 0.2:
                curdata.append(random.random() * 100)
                curindex.append(r)
        contents.append((numpy.array(curindex, dtype=itype), numpy.array(curdata, dtype=dtype)))
    return delayedarray.SparseNdarray((nr, nc), contents, dtype, itype)


def test_SparseNdarray_data_type():
    for dt in [numpy.float64, numpy.float32, numpy.int64, numpy.int32, numpy.int16, numpy.int8, numpy.uint64, numpy.uint32, numpy.uint16, numpy.uint8]:
        m = simulate_SparseNdarray(100, 50, dt, numpy.int32)
        ptr = initialize(m)
        assert all(ptr.row(0) == m[0, :])
        assert all(ptr.column(1) == m[:, 1])


def test_SparseNdarray_index_type():
    for it in [numpy.int64, numpy.int32, numpy.int16, numpy.int8, numpy.uint64, numpy.uint32, numpy.uint16, numpy.uint8]:
        m = simulate_SparseNdarray(50, 100, numpy.float64, it)
        ptr = initialize(m)
        assert all(ptr.row(0) == m[0, :])
        assert all(ptr.column(1) == m[:, 1])


def test_SparseNdarray_empty():
    m = delayedarray.SparseNdarray((10, 2), [None, None], numpy.float64, numpy.int32)
    ptr = initialize(m)
    assert all(ptr.row(0) == m[0, :])
    assert all(ptr.column(1) == m[:, 1])

    m = delayedarray.SparseNdarray((10, 2), None, numpy.float64, numpy.int32)
    ptr = initialize(m)
    assert all(ptr.row(0) == m[0, :])
    assert all(ptr.column(1) == m[:, 1])



"""Dummy conftest.py for mattress.

If you don't know what this is for, just leave it empty.
Read more about conftest.py under:
- https://docs.pytest.org/en/stable/fixture.html
- https://docs.pytest.org/en/stable/writing_plugins.html
"""

# import pytest



import numpy as np
import delayedarray as da
from mattress import initialize

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_delayed_subset():
    y = np.random.rand(1000, 100)
    x = da.DelayedArray(y)

    sub = x[:, 1:20:2]
    assert isinstance(sub.seed, da.Subset)
    ptr = initialize(sub)
    assert all(ptr.row(0) == y[0, 1:20:2])
    assert all(ptr.column(1) == y[:, 3])

    sub = x[5:200:5, :]
    assert isinstance(sub.seed, da.Subset)
    ptr = initialize(sub)
    assert all(ptr.row(0) == y[5, :])
    assert all(ptr.column(1) == y[5:200:5, 1])

    sub = x[5:200:5, 10:90:10]
    assert isinstance(sub.seed, da.Subset)
    ptr = initialize(sub)
    assert all(ptr.row(0) == y[5, 10:90:10])
    assert all(ptr.column(1) == y[5:200:5, 20])

    sub = x[
        np.ix_(np.array(range(5, 30, 2)), np.array(range(10, 30), dtype=np.int8))
    ]  # works with an existing numpy array.
    assert isinstance(sub.seed, da.Subset)
    ptr = initialize(sub)
    assert all(ptr.row(0) == y[5, 10:30])
    assert all(ptr.column(1) == y[5:30:2, 11])



from mattress import initialize
import numpy
import scipy.sparse

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_sparse_csr_matrix_dtype():
    m0 = scipy.sparse.rand(100, 50, density=0.25, format="csr", random_state=42) * 100
    for dt in [numpy.float64, numpy.float32, numpy.int64, numpy.int32, numpy.int16, numpy.int8, numpy.uint64, numpy.uint32, numpy.uint16, numpy.uint8]:
        m = m0.astype(dt)
        assert isinstance(m, scipy.sparse.csr_matrix)
        ptr = initialize(m)
        assert all(ptr.row(0) == m[0, :].toarray()[0])
        assert all(ptr.column(1) == m[:, 1].toarray().flatten())


def test_sparse_csr_matrix_index_dtype():
    m0 = scipy.sparse.rand(100, 50, density=0.25, format="csr", random_state=69)
    for it in [numpy.int64, numpy.int32, numpy.int16, numpy.int8, numpy.uint32, numpy.uint16, numpy.uint8]: # scipy doesn't support uint64, apparently...
        m = m0.copy()
        m.indices = m.indices.astype(it)
        assert isinstance(m, scipy.sparse.csr_matrix)
        ptr = initialize(m)
        assert all(ptr.row(0) == m[0, :].toarray()[0])
        assert all(ptr.column(1) == m[:, 1].toarray().flatten())


def test_sparse_csr_array():
    m = scipy.sparse.rand(3, 4, density=0.25, format="csr", random_state=42).tocsr()
    ptr = initialize(m)
    assert all(ptr.row(0) == m[0, :].toarray()[0])
    assert all(ptr.column(1) == m[:, 1].toarray().flatten())


def test_sparse_csc_matrix():
    m = scipy.sparse.rand(3, 4, density=0.25, format="csc", random_state=42)
    ptr = initialize(m)
    assert all(ptr.row(0) == m[0, :].toarray()[0])
    assert all(ptr.column(1) == m[:, 1].toarray().flatten())


def test_sparse_csc_array():
    m = scipy.sparse.rand(3, 4, density=0.25, format="csc", random_state=42).tocsc()
    ptr = initialize(m)
    assert all(ptr.row(0) == m[0, :].toarray()[0])
    assert all(ptr.column(1) == m[:, 1].toarray().flatten())



from functools import singledispatch
from typing import Any

import numpy
import delayedarray
from biocutils.package_utils import is_package_installed

from .InitializedMatrix import InitializedMatrix
from . import lib_mattress as lib
from ._utils import _sanitize_subset, _contiguify

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


@singledispatch
def initialize(x: Any) -> InitializedMatrix:
    """Initialize an :py:class:`~mattress.InitializedMatrix.InitializedMatrix`
    from a Python matrix representation. This prepares the matrix for use in
    C++ code that can accept a ``tatami::Matrix`` instance.

    Args:
        x: Any matrix-like object.

    Raises:
        NotImplementedError: if x is not supported.

    Returns:
        A pointer to tatami object.
    """
    raise NotImplementedError(
        f"initialize is not supported for objects of class: {type(x)}"
    )


@initialize.register
def _initialize_pointer(x: InitializedMatrix) -> InitializedMatrix:
    return x  # no-op


@initialize.register
def _initialize_numpy(x: numpy.ndarray) -> InitializedMatrix:
    if len(x.shape) != 2:
        raise ValueError("'x' should be a 2-dimensional array")
    x = _contiguify(x)
    return InitializedMatrix(
        ptr=lib.initialize_dense_matrix(x.shape[0], x.shape[1], x),
        objects=[x]
    )


if is_package_installed("scipy"):
    import scipy.sparse

    @initialize.register
    def _initialize_sparse_csr_array(x: scipy.sparse.csr_array) -> InitializedMatrix:
        dtmp = _contiguify(x.data)
        itmp = _contiguify(x.indices)
        indtmp = x.indptr.astype(numpy.uint64, copy=False, order="A")
        return InitializedMatrix(
            ptr=lib.initialize_compressed_sparse_matrix(x.shape[0], x.shape[1], dtmp, itmp, indtmp, True),
            objects=[dtmp, itmp, indtmp],
        )


    @initialize.register
    def _initialize_sparse_csr_matrix(x: scipy.sparse.csr_matrix) -> InitializedMatrix:
        return _initialize_sparse_csr_array(x)


    @initialize.register
    def _initialize_sparse_csc_array(x: scipy.sparse.csc_array) -> InitializedMatrix:
        dtmp = _contiguify(x.data)
        itmp = _contiguify(x.indices)
        indtmp = x.indptr.astype(numpy.uint64, copy=False, order="A")
        return InitializedMatrix(
            ptr=lib.initialize_compressed_sparse_matrix(x.shape[0], x.shape[1], dtmp, itmp, indtmp, False),
            objects=[dtmp, itmp, indtmp],
        )


    @initialize.register
    def _initialize_sparse_csc_matrix(x: scipy.sparse.csc_matrix) -> InitializedMatrix:
        return _initialize_sparse_csc_array(x)


@initialize.register
def _initialize_delayed_array(x: delayedarray.DelayedArray) -> InitializedMatrix:
    return initialize(x.seed)


@initialize.register
def _initialize_SparseNdarray(x: delayedarray.SparseNdarray) -> InitializedMatrix:
    if x.contents is not None:
        dvecs = []
        ivecs = []
        for y in x.contents:
            if y is None:
                ivecs.append(None)
                dvecs.append(None)
            else:
                ivecs.append(_contiguify(y[0]))
                dvecs.append(_contiguify(y[1]))
    else:
        nc = x.shape[1]
        dvecs = [None] * nc
        ivecs = [None] * nc

    return InitializedMatrix(
        ptr=lib.initialize_fragmented_sparse_matrix(x.shape[0], x.shape[1], dvecs, ivecs, False, x.dtype, x.index_dtype),
        objects=[dvecs, ivecs]
    )


@initialize.register
def _initialize_delayed_unary_isometric_operation_simple(
    x: delayedarray.UnaryIsometricOpSimple,
) -> InitializedMatrix:
    components = initialize(x.seed)
    ptr = lib.initialize_delayed_unary_isometric_operation_simple(
        components.ptr, x.operation.encode("UTF-8")
    )
    return InitializedMatrix(ptr, components.objects)


@initialize.register
def _initialize_delayed_unary_isometric_operation_with_args(
    x: delayedarray.UnaryIsometricOpWithArgs,
) -> InitializedMatrix:
    components = initialize(x.seed)
    obj = components.objects

    if isinstance(x.value, numpy.ndarray):
        contents = x.value.astype(numpy.float64, copy=False, order="A")
        ptr = lib.initialize_delayed_unary_isometric_operation_with_vector(
            components.ptr, x.operation.encode("UTF-8"), x.right, (x.along == 0), contents
        )
        obj.append(contents)
    else:
        ptr = lib.initialize_delayed_unary_isometric_operation_with_scalar(
            components.ptr, x.operation.encode("UTF-8"), x.right, x.value
        )

    return InitializedMatrix(ptr, obj)


@initialize.register
def _initialize_delayed_subset(
    x: delayedarray.Subset,
) -> InitializedMatrix:
    components = initialize(x.seed)

    for dim in range(2):
        current = x.subset[dim]
        noop, current = _sanitize_subset(current, x.shape[dim])
        if not noop:
            ptr = lib.initialize_delayed_subset(
                components.ptr, current, dim == 0
            )
            components = InitializedMatrix(ptr, components.objects + [current])

    return components


@initialize.register
def _initialize_delayed_bind(
    x: delayedarray.Combine,
) -> InitializedMatrix:
    collected = []
    objects = []
    for i, s in enumerate(x.seeds):
        components = initialize(s)
        collected.append(components.ptr)
        objects += components.objects

    ptr = lib.initialize_delayed_bind(collected, x.along)
    return InitializedMatrix(ptr, objects)


@initialize.register
def _initialize_delayed_transpose(
    x: delayedarray.Transpose,
) -> InitializedMatrix:
    components = initialize(x.seed)

    if x.perm == (1, 0):
        ptr = lib.initialize_delayed_transpose(components.ptr)
        components = InitializedMatrix(ptr, components.objects)

    return components


@initialize.register
def _initialize_delayed_binary_isometric_operation(
    x: delayedarray.BinaryIsometricOp,
) -> InitializedMatrix:
    lcomponents = initialize(x.left)
    rcomponents = initialize(x.right)

    ptr = lib.initialize_delayed_binary_isometric_operation(
        lcomponents.ptr, rcomponents.ptr, x.operation.encode("UTF-8")
    )

    return InitializedMatrix(ptr, lcomponents.objects + rcomponents.objects)


@initialize.register
def _initialize_delayed_round(
    x: delayedarray.Round,
) -> InitializedMatrix:
    components = initialize(x.seed)

    if x.decimals != 0:
        raise NotImplementedError(
            "non-zero decimals in 'delayedarray.Round' are not yet supported"
        )

    ptr = lib.initialize_delayed_unary_isometric_operation_simple(
        components.ptr, "round".encode("UTF-8")
    )

    return InitializedMatrix(ptr, components.objects)



import numpy as np
import delayedarray as da
from mattress import initialize

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_delayed_unary_isometric_log():
    y = np.random.rand(1000, 100)
    x = da.DelayedArray(y)

    z = np.log1p(x)
    assert isinstance(z.seed, da.UnaryIsometricOpSimple)
    ptr = initialize(z)
    assert np.allclose(ptr.row(0), np.log1p(y[0, :]))
    assert np.allclose(ptr.column(1), np.log1p(y[:, 1]))

    z = np.log2(x)
    ptr = initialize(z)
    assert np.allclose(ptr.row(0), np.log2(y[0, :]))
    assert np.allclose(ptr.column(1), np.log2(y[:, 1]))

    z = np.log10(x)
    ptr = initialize(z)
    assert np.allclose(ptr.row(0), np.log10(y[0, :]))
    assert np.allclose(ptr.column(1), np.log10(y[:, 1]))

    z = np.log(x)
    ptr = initialize(z)
    assert np.allclose(ptr.row(0), np.log(y[0, :]))
    assert np.allclose(ptr.column(1), np.log(y[:, 1]))


def test_delayed_unary_isometric_sign():
    y = np.random.rand(1000, 100) - 0.5
    x = da.DelayedArray(y)

    z = np.sign(x)
    assert isinstance(z.seed, da.UnaryIsometricOpSimple)
    ptr = initialize(z)
    assert all(ptr.row(0) == np.sign(y[0, :]))
    assert all(ptr.column(1) == np.sign(y[:, 1]))

    z = np.abs(x)
    ptr = initialize(z)
    assert all(ptr.row(0) == np.abs(y[0, :]))
    assert all(ptr.column(1) == np.abs(y[:, 1]))


def test_delayed_unary_isometric_sqrt():
    y = np.random.rand(1000, 100)
    x = da.DelayedArray(y)

    z = np.sqrt(x)
    assert isinstance(z.seed, da.UnaryIsometricOpSimple)
    ptr = initialize(z)
    assert np.allclose(ptr.row(0), np.sqrt(y[0, :]))
    assert np.allclose(ptr.column(1), np.sqrt(y[:, 1]))


def test_delayed_unary_isometric_int():
    y = np.random.rand(1000, 100) * 10
    x = da.DelayedArray(y)

    z = np.ceil(x)
    assert isinstance(z.seed, da.UnaryIsometricOpSimple)
    ptr = initialize(z)
    assert all(ptr.row(0) == np.ceil(y[0, :]))
    assert all(ptr.column(1) == np.ceil(y[:, 1]))

    z = np.floor(x)
    ptr = initialize(z)
    assert all(ptr.row(0) == np.floor(y[0, :]))
    assert all(ptr.column(1) == np.floor(y[:, 1]))

    z = np.trunc(x)
    ptr = initialize(z)
    assert all(ptr.row(0) == np.trunc(y[0, :]))
    assert all(ptr.column(1) == np.trunc(y[:, 1]))

    z = np.round(x)
    ptr = initialize(z)
    assert all(ptr.row(0) == np.round(y[0, :]))
    assert all(ptr.column(1) == np.round(y[:, 1]))


def test_delayed_unary_isometric_exp():
    y = np.random.rand(1000, 100)
    x = da.DelayedArray(y)

    z = np.exp(x)
    assert isinstance(z.seed, da.UnaryIsometricOpSimple)
    ptr = initialize(z)
    assert np.allclose(ptr.row(0), np.exp(y[0, :]))
    assert np.allclose(ptr.column(1), np.exp(y[:, 1]))

    z = np.expm1(x)
    ptr = initialize(z)
    assert np.allclose(ptr.row(0), np.expm1(y[0, :]))
    assert np.allclose(ptr.column(1), np.expm1(y[:, 1]))


def test_delayed_unary_isometric_trig():
    y = np.random.rand(1000, 100)
    x = da.DelayedArray(y)

    for op in [
        "cos",
        "sin",
        "tan",
        "cosh",
        "sinh",
        "tanh",
        "arccos",
        "arcsin",
        "arctan",
        "arctanh",
    ]:
        fun = getattr(np, op)
        z = fun(x)
        assert isinstance(z.seed, da.UnaryIsometricOpSimple)
        ptr = initialize(z)
        assert np.allclose(ptr.row(0), fun(y[0, :]))
        assert np.allclose(ptr.column(1), fun(y[:, 1]))

    y = y + 1
    x = da.DelayedArray(y)
    for op in ["arccosh", "arcsinh"]:
        fun = getattr(np, op)
        z = fun(x)
        assert isinstance(z.seed, da.UnaryIsometricOpSimple)
        ptr = initialize(z)
        assert np.allclose(ptr.row(0), fun(y[0, :]))
        assert np.allclose(ptr.column(1), fun(y[:, 1]))



import numpy as np
import delayedarray as da
from mattress import initialize

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_delayed_bind():
    y1 = np.random.rand(1000, 100)
    y2 = np.random.rand(1000, 50)

    com = np.concatenate((da.DelayedArray(y1), da.DelayedArray(y2)), axis=1)
    ref = np.concatenate((y1, y2), axis=1)
    assert isinstance(com.seed, da.Combine)
    ptr = initialize(com)
    assert all(ptr.row(0) == ref[0, :])
    assert all(ptr.column(1) == ref[:, 1])

    y1 = np.random.rand(1000, 10)
    y2 = np.random.rand(500, 10)
    y3 = np.random.rand(200, 10)

    com = np.concatenate(
        (da.DelayedArray(y1), da.DelayedArray(y2), da.DelayedArray(y3))
    )
    ref = np.concatenate((y1, y2, y3))
    assert isinstance(com.seed, da.Combine)
    ptr = initialize(com)
    assert all(ptr.row(0) == ref[0, :])
    assert all(ptr.column(1) == ref[:, 1])



import numpy as np
import delayedarray as da
from mattress import initialize

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_delayed_binary_isometric_arith():
    y1 = np.random.rand(1000, 100)
    y2 = np.random.rand(1000, 100)

    x = da.DelayedArray(y1) + da.DelayedArray(y2)
    assert isinstance(x.seed, da.BinaryIsometricOp)
    ptr = initialize(x)
    ref = y1 + y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) - da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 - y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) * da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 * y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) / da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 / y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) % da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 % y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) // da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 // y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) ** da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1**y2
    assert np.allclose(ptr.row(0), ref[0, :])
    assert np.allclose(ptr.column(1), ref[:, 1])


def test_delayed_binary_isometric_compare():
    y1 = np.random.rand(1000, 100)
    y2 = np.random.rand(1000, 100)

    x = da.DelayedArray(y1) == da.DelayedArray(y2)
    assert isinstance(x.seed, da.BinaryIsometricOp)
    ptr = initialize(x)
    ref = y1 == y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) != da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 != y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) > da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 > y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) >= da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 >= y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) < da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 < y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = da.DelayedArray(y1) <= da.DelayedArray(y2)
    ptr = initialize(x)
    ref = y1 <= y2
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()


def test_delayed_binary_isometric_logical():
    y1 = np.random.rand(1000, 100)
    y2 = np.random.rand(1000, 100)

    x = np.logical_or(da.DelayedArray(y1), da.DelayedArray(y2))
    assert isinstance(x.seed, da.BinaryIsometricOp)
    ptr = initialize(x)
    ref = np.logical_or(y1, y2)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = np.logical_and(da.DelayedArray(y1), da.DelayedArray(y2))
    ptr = initialize(x)
    ref = np.logical_and(y1, y2)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()

    x = np.logical_xor(da.DelayedArray(y1), da.DelayedArray(y2))
    ptr = initialize(x)
    ref = np.logical_xor(y1, y2)
    assert (ptr.row(0) == ref[0, :]).all()
    assert (ptr.column(1) == ref[:, 1]).all()



import os
from typing import Union, Tuple, Sequence, List
import inspect
import numpy as np

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


def _sanitize_subset(
    subset: Sequence[int], full: int
) -> Tuple[bool, Union[np.ndarray, None]]:
    is_noop = True
    if len(subset) == full:
        for i, x in enumerate(subset):
            if i != x:
                is_noop = False
                break
    else:
        is_noop = False

    if is_noop:
        return True, np.ndarray(0, dtype=np.uint32)

    if not isinstance(subset, np.ndarray):
        subset = np.array(subset, dtype=np.uint32)
    else:
        subset = subset.astype(
            np.uint32, copy=not (subset.flags.c_contiguous or subset.flags.f_contiguous)
        )

    return False, subset


def _contiguify(x: np.ndarray):
    if x.flags["C_CONTIGUOUS"] or x.flags["F_CONTIGUOUS"]:
        return x
    else:
        return x.copy('C')



import numpy
import delayedarray
from . import lib_mattress as lib
from typing import Tuple, Sequence
from ._utils import _sanitize_subset

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def _factorize(group):
    mapping = {}
    indices = numpy.ndarray((len(group),), numpy.uint32)
    levels = []
    for i, x in enumerate(group):
        if x not in mapping:
            mapping[x] = len(levels)
            levels.append(x)
        indices[i] = mapping[x]
    return levels, indices


class InitializedMatrix:
    """Pointer to an initialized ``tatami::matrix``, for use in C++ code.
    Instances of this class should only be created by developers and used
    within package functions; this is done by fetching the :py:attr:`~ptr`
    attribute and passing it as a ``std::shared_ptr<tatami::Matrix<double,
    uint32_t> >`` in C++ code, e.g., via pybind11. All ``InitializedMatrix``
    instances are expected to be transient within a Python session; they should
    not be serialized, nor should they be visible to end users. Each instance
    will automatically free the C++-allocated memory upon garbage collection.
    """

    def __init__(self, ptr, objects: list):
        """
        Args:
            ptr:
                Shared pointer to a ``tatami::Matrix<double, uint32_t>``
                instance, created and wrapped by pybind11.

            objects:
                List of Python objects (typically NumPy arrays) to protect from
                garbage collection, as their data is referenced by ``ptr``.
        """
        self._ptr = ptr
        self._objects = objects

    def nrow(self) -> int:
        """Get number of rows.

        Returns:
            Number of rows.
        """
        return self.shape[0];

    def ncol(self) -> int:
        """Get number of columns.

        Returns:
            Number of columns.
        """
        return self.shape[1];

    @property
    def shape(self) -> Tuple[int, int]:
        """Shape of the matrix, to masquerade as a NumPy-like object."""
        return lib.extract_dim(self._ptr)

    @property
    def ptr(self): 
        """Shared pointer to a ``tatami::Matrix<double, uint32_t>`` instance,
        to be passed to C++ code via pybind11."""
        return self._ptr

    @property
    def objects(self) -> list:
        """List of objects to protect from garbage collection as they are
        referenced by :py:attr:`~ptr`."""
        return self._objects

    @property
    def dtype(self) -> numpy.dtype:
        """Type of the matrix, to masquerade as a NumPy-like object."""
        return numpy.dtype("float64")

    def __array__(self) -> numpy.ndarray:
        """Realize the underlying matrix into a dense NumPy array.

        Returns:
            Contents of the underlying matrix.
        """
        shape = self.shape;
        return _extract_array(self._ptr, (range(shape[0]), range(shape[1])), sparse=False)

    def __DelayedArray_dask__(self) -> numpy.ndarray:
        """Enable the use of the poiners with dask.

        See :py:meth:`~delayedarray.DelayedArray.DelayedArray.__DelayedArray_dask__` for details.

        This is largely a placeholder method for compatibility; it just realizes the underlying matrix into a dense
        array under the hood.

        Returns:
            Contents of the underlying matrix.
        """
        return self.__array___()

    def sparse(self) -> bool:
        """Is the matrix sparse?

        Returns:
            True if matrix is sparse.
        """
        return lib.extract_sparse(self._ptr); 

    def row(self, r: int) -> numpy.ndarray:
        """Access a row from the tatami matrix. This method is primarily intended for troubleshooting and should not be
        used to iterate over the matrix in production code. (Do that in C++ instead.)

        Args:
            r: Row to access.

        Returns:
            Row from the matrix. This is always in double-precision,
            regardless of the underlying representation.
        """
        return lib.extract_row(self._ptr, r)

    def column(self, c: int) -> numpy.ndarray:
        """Access a column from the tatami matrix. This method is primarily intended for troubleshooting and should not
        be used to iterate over the matrix in production code. (Do that in C++ instead.)

        Args:
            c: Column to access.

        Returns:
            Column from the matrix. This is always in double-precision,
            regardless of the underlying representation.
        """
        return lib.extract_column(self._ptr, c)

    def row_sums(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute row sums.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row sums.
        """
        return lib.compute_row_sums(self._ptr, num_threads)

    def column_sums(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute column sums.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column sums.
        """
        return lib.compute_column_sums(self._ptr, num_threads)

    def row_variances(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute row variances.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row variances.
        """
        return lib.compute_row_variances(self._ptr, num_threads)

    def column_variances(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute column variances.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column variances.
        """
        return lib.compute_column_variances(self._ptr, num_threads)

    def row_medians(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute row medians.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row medians.
        """
        return lib.compute_row_medians(self._ptr, num_threads)

    def column_medians(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute column medians.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column medians.
        """
        return lib.compute_column_medians(self._ptr, num_threads)

    def row_mins(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute row minima.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row minima.
        """
        return lib.compute_row_mins(self._ptr, num_threads)

    def column_mins(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute column minima.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column minima.
        """
        return lib.compute_column_mins(self._ptr, num_threads)

    def row_maxs(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute row maxima.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row maxima.
        """
        return lib.compute_row_maxs(self._ptr, num_threads)

    def column_maxs(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to compute column maxima.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column maxima.
        """
        return lib.compute_column_maxs(self._ptr, num_threads)

    def row_ranges(self, num_threads: int = 1) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """Convenience method to compute row ranges.

        Args:
            num_threads: Number of threads.

        Returns:
            Tuple containing the row minima and maxima.
        """
        return lib.compute_row_ranges(self._ptr, num_threads)

    def column_ranges(self, num_threads: int = 1) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """Convenience method to compute column ranges.

        Args:
            num_threads: Number of threads.

        Returns:
            Tuple containing the column minima and maxima.
        """
        return lib.compute_column_ranges(self._ptr, num_threads)

    def row_nan_counts(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to count the number of NaNs on each row.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of row NaN counts.
        """
        return lib.compute_row_nan_counts(self._ptr, num_threads)

    def column_nan_counts(self, num_threads: int = 1) -> numpy.ndarray:
        """Convenience method to count the number of NaNs on each column.

        Args:
            num_threads: Number of threads.

        Returns:
            Array of column NaN counts.
        """
        return lib.compute_column_nan_counts(self._ptr, num_threads)

    def row_medians_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[numpy.ndarray, list]:
        """Convenience method to compute the row-wise median for each group of columns.

        Args:
            group: Sequence of length equal to the number of columns of the matrix,
                containing the group assignment for each column.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each column represents
            a group and contains the row-wise medians for that group, across all rows of the matrix;
            and a list containing the unique levels of ``group`` represented by each column.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.ncol():
            raise ValueError(
                "'group' should have length equal to the number of columns"
            )

        output = lib.compute_row_medians_by_group(self._ptr, ind, num_threads);
        return output, lev

    def column_medians_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[numpy.ndarray, list]:
        """Convenience method to compute the column-wise median for each group of row.

        Args:
            group: Sequence of length equal to the number of row of the matrix,
                containing the group assignment for each row.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each row represents a
            group and contains the column-wise medians for that group, across
            all columns of the matrix; and a list containing the unique levels
            of ``group`` represented by each row.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.nrow():
            raise ValueError("'group' should have length equal to the number of rows")

        output = lib.compute_column_medians_by_group(self._ptr, ind, num_threads)
        return output.T, lev

    def row_sums_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[numpy.ndarray, list]:
        """Convenience method to compute the row-wise sum for each group of columns.

        Args:
            group: Sequence of length equal to the number of columns of the matrix,
                containing the group assignment for each column.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each column represents
            a group and contains the row-wise sums for that group, across all
            rows of the matrix; and a list containing the unique levels of
            ``group`` represented by each column.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.ncol():
            raise ValueError(
                "'group' should have length equal to the number of columns"
            )

        output = lib.compute_row_sums_by_group(self._ptr, ind, num_threads)
        return output, lev

    def column_sums_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[numpy.ndarray, list]:
        """Convenience method to compute the column-wise sum for each group of row.

        Args:
            group: Sequence of length equal to the number of row of the matrix,
                containing the group assignment for each row.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each row represents a
            group and contains the column-wise sums for that group, across all
            columns of the matrix; and a list containing the unique levels of
            ``group`` represented by each row.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.nrow():
            raise ValueError("'group' should have length equal to the number of rows")

        output = lib.compute_column_sums_by_group(self._ptr, ind, num_threads)
        return output.T, lev

    def row_variances_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[numpy.ndarray, list]:
        """Convenience method to compute the row-wise variance for each group of columns.

        Args:
            group: Sequence of length equal to the number of columns of the matrix,
                containing the group assignment for each column.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each column represents
            a group and contains the row-wise variances for that group, across all
            rows of the matrix; and a list containing the unique levels of
            ``group`` represented by each column.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.ncol():
            raise ValueError(
                "'group' should have length equal to the number of columns"
            )

        output = lib.compute_row_variances_by_group(self._ptr, ind, num_threads)
        return output, lev

    def column_variances_by_group(
        self, group: Sequence, num_threads: int = 1
    ) -> Tuple[numpy.ndarray, list]:
        """Convenience method to compute the column-wise variance for each group of row.

        Args:
            group: Sequence of length equal to the number of row of the matrix,
                containing the group assignment for each row.

            num_threads: Number of threads.

        Returns:
            Tuple containing a 2-dimensional array where each row represents a
            group and contains the column-wise variances for that group, across all
            columns of the matrix; and a list containing the unique levels of
            ``group`` represented by each row.
        """
        lev, ind = _factorize(group)
        if len(ind) != self.nrow():
            raise ValueError("'group' should have length equal to the number of rows")

        output = lib.compute_column_variances_by_group(self._ptr, ind, num_threads)
        return output.T, lev



@delayedarray.is_sparse.register
def is_sparse(x: InitializedMatrix):
    """See :py:func:`~delayedarray.is_sparse.is_sparse`."""
    return x.sparse()


def _extract_array(x: InitializedMatrix, subset: Tuple[Sequence[int], ...], sparse: bool): 
    shape = x.shape
    rnoop, rsub = _sanitize_subset(subset[0], shape[0])
    cnoop, csub = _sanitize_subset(subset[1], shape[1])
    if not sparse:
        return lib.extract_dense_subset(x.ptr, rnoop, rsub, cnoop, csub)
    else:
        return lib.extract_sparse_subset(x.ptr, rnoop, rsub, cnoop, csub)


@delayedarray.extract_dense_array.register
def extract_dense_array(x: InitializedMatrix, subset: Tuple[Sequence[int], ...]) -> numpy.ndarray:
    """See :py:func:`~delayedarray.extract_dense_array.extract_dense_array`."""
    return _extract_array(x, subset, False)

@delayedarray.extract_sparse_array.register
def extract_sparse_array(x: InitializedMatrix, subset: Tuple[Sequence[int], ...]) -> delayedarray.SparseNdarray:
    """See :py:func:`~delayedarray.extract_sparse_array.extract_sparse_array`."""
    return _extract_array(x, subset, True)

@delayedarray.is_masked.register
def is_masked(x: InitializedMatrix) -> bool:
    """See :py:func:`~delayedarray.is_masked.is_masked`."""
    return False

@delayedarray.chunk_grid.register
def chunk_grid(x: InitializedMatrix) -> bool:
    """See :py:func:`~delayedarray.chunk_grid.chunk_grid`."""
    return delayedarray.chunk_shape_to_grid((1, 1), x.shape, cost_factor=1)



import numpy as np
import delayedarray as da
from mattress import initialize

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_delayed_transpose():
    y = np.random.rand(1000, 100)
    x = da.DelayedArray(y)

    t = x.T
    ptr = initialize(t)
    assert all(ptr.row(0) == y[:, 0])
    assert all(ptr.column(1) == y[1, :])

    noop = np.transpose(x, axes=(0, 1))
    x2 = da.DelayedArray(noop)
    ptr = initialize(x2)
    assert all(ptr.row(0) == y[0, :])
    assert all(ptr.column(1) == y[:, 1])



"""Setup file for mattress. Use setup.cfg to configure your project.

This file was generated with PyScaffold 4.5.
PyScaffold helps you to put up the scaffold of your new Python project.
Learn more under: https://pyscaffold.org/
"""
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as build_ext_orig
from glob import glob
import pathlib
import os
import shutil
import sys
import pybind11

## Adapted from https://stackoverflow.com/questions/42585210/extending-setuptools-extension-to-use-cmake-in-setup-py.
class CMakeExtension(Extension):
    def __init__(self, name):
        super().__init__(name, sources=[])

class build_ext(build_ext_orig):
    def run(self):
        for ext in self.extensions:
            self.build_cmake(ext)

    def build_cmake(self, ext):
        build_temp = pathlib.Path(self.build_temp)
        build_lib = pathlib.Path(self.build_lib)
        outpath = os.path.join(build_lib.absolute(), ext.name) 

        if not os.path.exists(build_temp):
            import assorthead
            cmd = [ 
                "cmake", 
                "-S", "lib",
                "-B", build_temp,
                "-Dpybind11_DIR=" + os.path.join(os.path.dirname(pybind11.__file__), "share", "cmake", "pybind11"),
                "-DPYTHON_EXECUTABLE=" + sys.executable,
                "-DASSORTHEAD_INCLUDE_DIR=" + assorthead.includes()
            ]
            if os.name != "nt":
                cmd.append("-DCMAKE_BUILD_TYPE=Release")
                cmd.append("-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + outpath)

            if "MORE_CMAKE_OPTIONS" in os.environ:
                cmd += os.environ["MORE_CMAKE_OPTIONS"].split()
            self.spawn(cmd)

        if not self.dry_run:
            cmd = ['cmake', '--build', build_temp]
            if os.name == "nt":
                cmd += ["--config", "Release"]
            self.spawn(cmd)
            if os.name == "nt": 
                # Gave up trying to get MSVC to respect the output directory.
                # Delvewheel also needs it to have a 'pyd' suffix... whatever.
                shutil.copyfile(os.path.join(build_temp, "Release", "_core.dll"), os.path.join(outpath, "_core.pyd"))


if __name__ == "__main__":
    import os
    try:
        setup(
            use_scm_version={"version_scheme": "no-guess-dev"},
            ext_modules=[CMakeExtension("mattress")],
            cmdclass={
                'build_ext': build_ext
            }
        )
    except:  # noqa
        print(
            "\n\nAn error occurred while building the project, "
            "please ensure you have the most updated version of setuptools, "
            "setuptools_scm and wheel with:\n"
            "   pip install -U setuptools setuptools_scm wheel\n\n"
        )
        raise



import numpy as np
from mattress import initialize, InitializedMatrix
import delayedarray as da
import scipy

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


def test_pointer_rewrap():
    y = np.random.rand(1000, 100)
    ptr = initialize(y)
    reptr = initialize(ptr)
    assert ptr.ptr == reptr.ptr


def test_simple_stats():
    y = np.random.rand(1000, 100)
    ptr = initialize(y)

    assert np.allclose(ptr.row_sums(), y.sum(axis=1))
    assert np.allclose(ptr.column_sums(), y.sum(axis=0))
    assert np.allclose(ptr.row_variances(), y.var(axis=1, ddof=1))
    assert np.allclose(ptr.column_variances(), y.var(axis=0, ddof=1))

    assert ptr.row_medians().shape == (1000,)
    assert ptr.column_medians().shape == (100,)

    assert (ptr.row_mins() == y.min(axis=1)).all()
    assert (ptr.column_mins() == y.min(axis=0)).all()
    assert (ptr.row_maxs() == y.max(axis=1)).all()
    assert (ptr.column_maxs() == y.max(axis=0)).all()

    mn, mx = ptr.row_ranges()
    assert (mn == y.min(axis=1)).all()
    assert (mx == y.max(axis=1)).all()
    mn, mx = ptr.column_ranges()
    assert (mn == y.min(axis=0)).all()
    assert (mx == y.max(axis=0)).all()


def test_nan_counts():
    y = np.random.rand(20, 10)
    y[0, 0] = np.nan
    y[1, 2] = np.nan
    y[4, 3] = np.nan
    y[2, 3] = np.nan

    ptr = initialize(y)
    rnan = ptr.row_nan_counts()
    cnan = ptr.column_nan_counts()
    assert rnan[0] == 1
    assert cnan[3] == 2


def test_grouped_stats():
    y = np.random.rand(20, 10)
    ptr = initialize(y)

    cgrouping = ["D", "C", "B", "A", "A", "B", "C", "B", "A", "A"]
    rgrouping = []
    for i in range(20):
        rgrouping.append(i % 3)

    rmed, rlev = ptr.row_medians_by_group(cgrouping)
    assert rmed.shape == (20, 4)
    assert len(rlev) == 4
    assert (rmed[:, rlev.index("D")] == y[:, 0]).all()

    cmed, clev = ptr.column_medians_by_group(rgrouping)
    assert cmed.shape == (3, 10)
    assert len(clev) == 3

    keep = []
    for i, x in enumerate(rgrouping):
        if x == 1:
            keep.append(i)
    ref = y[keep, :]
    rptr = initialize(ref)
    assert (cmed[clev.index(1), :] == rptr.column_medians()).all()

    rsum, rlev = ptr.row_sums_by_group(cgrouping)
    assert rsum.shape == (20, 4)
    assert len(rlev) == 4
    assert (rsum[:, rlev.index("D")] == y[:, 0]).all()

    csum, clev = ptr.column_sums_by_group(rgrouping)
    assert csum.shape == (3, 10)
    assert len(clev) == 3

    keep = []
    for i, x in enumerate(rgrouping):
        if x == 0:
            keep.append(i)
    ref = y[keep, :]
    rptr = initialize(ref)
    assert (csum[clev.index(0), :] == rptr.column_sums()).all()


def test_DelayedArray_rewrap_dense():
    y = np.random.rand(1000, 100)
    ptr = initialize(y)

    da2 = da.DelayedArray(ptr)
    assert isinstance(da2.seed, InitializedMatrix)
    assert (np.array(da2) == y).all()

    # These all indirectly call extract_dense_array().
    sub = da2[10:50, 0:100:2]
    assert (np.array(sub) == y[10:50, 0:100:2]).all()

    sub = da2[10:50, :]
    assert (np.array(sub) == y[10:50, :]).all()

    sub = da2[:, 10:50]
    assert (np.array(sub) == y[:, 10:50]).all()


def test_DelayedArray_rewrap_sparse_csc():
    y = scipy.sparse.random(200, 500, 0.1).tocsc()
    ptr = initialize(y)

    da2 = da.DelayedArray(ptr)
    assert isinstance(da2.seed, InitializedMatrix)
    assert da.is_sparse(da2)

    full = da.to_sparse_array(da2)
    assert isinstance(full, da.SparseNdarray)
    assert (np.array(full) == y.toarray()).all()

    # These all indirectly call InitializedMatrix's extract_dense_array().
    sub = da.extract_sparse_array(da2, (range(10, 50), range(0, 100, 2)))
    assert (np.array(sub) == y[10:50, 0:100:2].toarray()).all()

    sub = da.extract_sparse_array(da2, (range(10, 50), range(y.shape[1])))
    assert (np.array(sub) == y[10:50, :].toarray()).all()

    sub = da.extract_sparse_array(da2, (range(y.shape[0]), range(10, 50)))
    assert (np.array(sub) == y[:, 10:50].toarray()).all()


def test_DelayedArray_rewrap_sparse_csr():
    y = scipy.sparse.random(200, 500, 0.1).tocsr()
    ptr = initialize(y)

    da2 = da.DelayedArray(ptr)
    assert isinstance(da2.seed, InitializedMatrix)
    assert da.is_sparse(da2)

    full = da.to_sparse_array(da2)
    assert isinstance(full, da.SparseNdarray)
    assert (np.array(full) == y.toarray()).all()

    # These all indirectly call InitializedMatrix's extract_dense_array().
    sub = da.extract_sparse_array(da2, (range(10, 50), range(0, 100, 2)))
    assert (np.array(sub) == y[10:50, 0:100:2].toarray()).all()

    sub = da.extract_sparse_array(da2, (range(10, 50), range(y.shape[1])))
    assert (np.array(sub) == y[10:50, :].toarray()).all()

    sub = da.extract_sparse_array(da2, (range(y.shape[0]), range(10, 50)))
    assert (np.array(sub) == y[:, 10:50].toarray()).all()
